package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckOnlyMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for executing an explicit
 * check of a containment constraint. It can be executed
 * once mappings have been created for both the container
 * and the contained node.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMContainmentCheckMatchingAction extends MLSDMCheckMatchingAction {

	/*
	 * Containment link to match.
	 */
	private AbstractStoryPatternLink link;
	
	/*
	 * Container node.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containingNode;
	
	/*
	 * Contained node.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containedNode;

	public MLSDMContainmentCheckMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, AbstractStoryPatternLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containingNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containedNode) {
		super(id, matcher);
		this.link = link;
		this.containingNode = containingNode;
		this.containedNode = containedNode;
		this.addRequiredBinding(containingNode);
		this.addRequiredBinding(containedNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		return new MLSDMCheckOnceMatchState();
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!containingNode.isBound() || !containedNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else{
			return PatternConstraint.CHECK_ONLY;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		
		if(!(matchState instanceof MLSDMCheckOnceMatchState)) {
			transaction.setValid(false);
		}
		else {
			MLSDMCheckOnceMatchState ms = (MLSDMCheckOnceMatchState) matchState;
			
			transaction.setValid(ms.check() && checkContainment());
		}
		return transaction;
	}

	/**
	 * Perform an explicit check of the containment constraint.
	 * 
	 * @return 	True if the contained node is (transitively) contained
	 * 			in the container node.
	 * 			
	 */
	private boolean checkContainment() {
		EObject containingInstance = (EObject) matcher.getVariable(containingNode.getSpo().getName()).getValue();
		EObject containedInstance = (EObject) matcher.getVariable(containedNode.getSpo().getName()).getValue();
		
		EObject currentContainer = containedInstance;
		while((currentContainer = currentContainer.eContainer()) != null) {
			if(currentContainer == containingInstance) {
				matcher.getNotificationEmitter().linkCheckSuccessful(link.getSource(), containingNode.getSpo(), link, link.getTarget(), containedNode.getSpo(), matcher.getVariablesScope(), matcher);
				return true;
			}
		}
		
		matcher.getNotificationEmitter().linkCheckFailed(link.getSource(), containingNode.getSpo(), link, link.getTarget(), containedNode.getSpo(), matcher.getVariablesScope(), matcher);
		return false;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction#check()
	 */
	@Override
	public boolean check() {
		return checkContainment();
	}

	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "check \t\t (" + link.getSource().getName() + " -*contains*-> " + link.getTarget().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMContainmentCheckMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(containingNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(containedNode));
	}

}
