package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Iterator;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching the contained node
 * of a containment constraint. It can be executed once the
 * a mapping for the container node has been created.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMContainmentMatchingAction extends MLSDMMatchingAction {

	/**
	 * This is a match state specifically for a containment
	 * matching action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private static class MLSDMContainmentMatchingActionMatchState extends MatchState{
		
		/*
		 * Iterator for enumerating candidate mappings for
		 * the contained node.
		 */
		private Iterator<EObject> iterator;

		private MLSDMContainmentMatchingActionMatchState(Iterator<EObject> iterator) {
			this.iterator = iterator;
		}

		private Iterator<EObject> getIterator() {
			return iterator;
		}
		
	}
	
	/*
	 * Container node.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containingNode;
	
	/*
	 * Contained node.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containedNode;
	
	/*
	 * Containment link to be matched.
	 */
	private AbstractStoryPatternLink link;
	
	public MLSDMContainmentMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, AbstractStoryPatternLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containingNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> containedNode) {
		super(id, matcher);
		this.link = link;
		this.containingNode = containingNode;
		this.containedNode = containedNode;
		this.addRequiredBinding(containingNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		AbstractStoryPatternObject containingSPO = containingNode.getSpo();
		
		if(!containingNode.isBound() || containedNode.isBound()) {
			return null;
		}
		else{
			EObject containingInstance = (EObject) matcher.getVariable(containingSPO.getName()).getValue();
			return new MLSDMContainmentMatchingActionMatchState(containingInstance.eAllContents());
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!containingNode.isBound() || containedNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else{
			return PatternConstraint.ALWAYS_MATCH_LAST;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
		
		EObject containingInstance = (EObject) matcher.getVariable(containingNode.getSpo().getName()).getValue();
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), containingInstance, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		if(!(matchState instanceof MLSDMContainmentMatchingActionMatchState)) {
			transaction.setValid(false);
		}
		else {
			boolean valid = false;
			MLSDMContainmentMatchingActionMatchState ms = (MLSDMContainmentMatchingActionMatchState) matchState;
			Iterator<? extends Object> featureIterator = ms.getIterator();
			Object containedInstance = null;
			while(!valid && featureIterator.hasNext()) {
				containedInstance = featureIterator.next();
				valid = matcher.instanceMatchesSPO(containedInstance, containedNode.getSpo()) && matcher.checkMatchingHistory(this.getPatternConstraint(), containedInstance);
			}
			
			if(valid) {
				transaction.setTargetInstance(containedInstance);
				transaction.setSPO(containedNode.getSpo());
			}
			else {
				matcher.getNotificationEmitter().storyPatternObjectNotBound(containedNode.getSpo(), matcher.getVariablesScope(),
						matcher);
			}
			
			transaction.setValid(valid);
			transaction.addPatternNode(containedNode);

			matcher.updateMatchingHistory(this.getPatternConstraint(), containedInstance);
		}
		return transaction;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "traverse \t (" + link.getTarget().getName() + " -*contains*-> " + link.getSource().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMContainmentMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(containingNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(containedNode));
	}
}
