package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Collection;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.ExpressionLink;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckOnlyMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for executing an explicit
 * check of a constraint based on an expression link. It
 * can be executed once mappings have been created for
 * both source and target node of the link.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMExpressionLinkCheckMatchingAction extends MLSDMCheckMatchingAction {

	/*
	 * Expression link.
	 */
	private ExpressionLink link;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Target node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode;
	
	public MLSDMExpressionLinkCheckMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, ExpressionLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode) {
		super(id, matcher);
		this.link = link;
		this.sourceNode = sourceNode;
		this.targetNode = targetNode;
		this.addRequiredBinding(sourceNode);
		this.addRequiredBinding(targetNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		if(!sourceNode.isBound() || !targetNode.isBound()) {
			return null;
		}
		else{
			return new MLSDMCheckOnceMatchState();
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!sourceNode.isBound() || !targetNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else{
			return PatternConstraint.CHECK_ONLY;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		if(!(matchState instanceof MLSDMCheckOnceMatchState)) {
			transaction.setValid(false);
		}
		else {
			transaction.setValid(((MLSDMCheckOnceMatchState) matchState).check() && checkLink());
		}
		return transaction;
	}

	/**
	 * Perform the actual link check by evaluating the
	 * expression associated with the link and checking
	 * if the result matches the existing mapping for the
	 * target node.
	 * 
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private boolean checkLink() {
		Object sourceInstance = matcher.getVariable(sourceNode.getSpo().getName()).getValue();
		Object targetInstance = matcher.getVariable(targetNode.getSpo().getName()).getValue();
		Object evaluationResult = null;
		try {
			evaluationResult = matcher.getExpressionInterpreterManager().evaluateExpression(link.getExpression(), sourceNode.getSpo().getType(), sourceInstance, matcher.getVariablesScope()).getValue();
		} catch (SDMException e) {
			// TODO throw this exception instead
			e.printStackTrace();
		}
		
		boolean success;
		
		if(evaluationResult instanceof Collection) {
			success = ((Collection<? extends Object>) evaluationResult).contains(targetInstance);
		}
		else {
			success = evaluationResult == targetInstance;
		}
		
		if(success) {
			matcher.getNotificationEmitter().linkCheckSuccessful(sourceNode.getSpo(), sourceInstance, link, targetNode.getSpo(), targetInstance, matcher.getVariablesScope(), matcher);
		}
		else {
			matcher.getNotificationEmitter().linkCheckFailed(sourceNode.getSpo(), sourceInstance, link, targetNode.getSpo(), targetInstance, matcher.getVariablesScope(), matcher);
		}
		
		return success;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction#check()
	 */
	@Override
	public boolean check() {
		return checkLink();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "check \t\t (" + link.getSource().getName() + " -" + (link.getExpression() instanceof MLStringExpression ? ((MLStringExpression) link.getExpression()).getExpressionString() : link.getExpression().toString()) + "-> " + link.getTarget().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMExpressionLinkCheckMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(targetNode));
	}

}
