package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Collection;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.ExpressionLink;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching the
 * target node of an expression link. It can be
 * executed once a mapping for the source node
 * of the link has been created.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMExpressionLinkTraverseMatchingAction extends MLSDMMatchingAction {

	/**
	 * This is a match state specific to an
	 * expression link traversal matching action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private class MLSDMTraverseExpressionLinkMatchState extends MatchState{
		
	}

	/*
	 * The expression link.
	 */
	private ExpressionLink link;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Target node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode;
	
	public MLSDMExpressionLinkTraverseMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, ExpressionLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode) {
		super(id, matcher);
		this.link = link;
		this.sourceNode = sourceNode;
		this.targetNode = targetNode;
		this.addRequiredBinding(sourceNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		if(!sourceNode.isBound() || targetNode.isBound()) {
			return null;
		}
		else{
			return new MLSDMTraverseExpressionLinkMatchState();
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!sourceNode.isBound() || targetNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else{
			return 100; //TODO provide a more useful estimate
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
		
		Object sourceInstance = matcher.getVariable(sourceNode.getSpo().getName()).getValue();
		
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), sourceInstance, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		Object evaluationResult = null;
		try {
			evaluationResult = matcher.getExpressionInterpreterManager().evaluateExpression(link.getExpression(), sourceNode.getSpo().getType(), sourceInstance, matcher.getVariablesScope()).getValue();
		} catch (SDMException e) {
			// TODO throw this exception instead
			e.printStackTrace();
		}
		
		if(evaluationResult instanceof Collection) {
			/*
			 * Check whether an object in the collection
			 * is a valid mapping for the target node.
			 */
			for(Object targetInstance:(Collection<? extends Object>) evaluationResult) {
				boolean valid = matcher.instanceMatchesSPO(targetInstance, targetNode.getSpo()) && matcher.checkMatchingHistory(getPatternConstraint(), targetInstance);
				matcher.updateMatchingHistory(getPatternConstraint(), targetInstance);
			
				if(valid) {
					transaction.setValid(true);
					transaction.setSPO(targetNode.getSpo());
					transaction.setTargetInstance(targetInstance);
					
					transaction.addPatternNode(targetNode);
					return transaction;
				}
				
			}
		}
		else {
			/*
			 * Check whether the result of the evaluation
			 * is a valid mapping for the target node.
			 */
			Object targetInstance = evaluationResult;
			
			boolean valid = matcher.instanceMatchesSPO(targetInstance, targetNode.getSpo()) && matcher.checkMatchingHistory(getPatternConstraint(), targetInstance);
			matcher.updateMatchingHistory(getPatternConstraint(), targetInstance);
		
			if(valid) {
				transaction.setValid(true);
				transaction.setSPO(targetNode.getSpo());
				transaction.setTargetInstance(targetInstance);
				
				transaction.addPatternNode(targetNode);
				return transaction;
			}
		}
		
		matcher.getNotificationEmitter().storyPatternObjectNotBound(targetNode.getSpo(), matcher.getVariablesScope(),
				matcher);
		
		transaction.setValid(false);
		return transaction;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "traverse \t (" + link.getSource().getName() + " -" + (link.getExpression() instanceof MLStringExpression ? ((MLStringExpression) link.getExpression()).getExpressionString() : link.getExpression().toString()) + "-> " + link.getTarget().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMExpressionLinkTraverseMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(targetNode));
	}

}
