package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlsdm.mlindices.Index;
import de.mdelab.mlsdm.mlindices.IndexEntry;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLPackage;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLRange;
import de.mdelab.sdm.icl.iCL.ICLValue;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchMultipleNodesMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching only the node
 * referenced by the last parameter in the query template
 * of an index constraint. It can be executed once mappings
 * have been created for all other dependencies of the
 * index constraint.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMIndexLastParameterMatchingAction extends MLSDMIndexMatchingAction {

	public MLSDMIndexLastParameterMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, ICLConstraint iclConstraint, MLStringExpression expression,
			Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode) {
		super(id, matcher, iclConstraint, expression, iclValueToPatternNode);
		this.iclConstraint = iclConstraint;
		this.expression = expression;
		this.index = (Index) matcher.getVariable(iclConstraint.getIndexID()).getValue();
		initializeRequirements();
	}

	private void initializeRequirements() {
		for(int i = 0; i < iclConstraint.getParameters().getList().size() - 1; i++) {
			ICLParameter parameter = iclConstraint.getParameters().getList().get(i);
			/*
			 * all parameters except for the last one have to be bound
			 * as this matching action should only match the last parameter
			 */
			if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
				Collection<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> mins
									= iclValueToPatternNode.get(((ICLRange)parameter).getMin());
				if(mins != null) {
					for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> min:mins) {
						addRequiredBinding(min);
					}
				}
				
				Collection<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> maxs
						= iclValueToPatternNode.get(((ICLRange)parameter).getMin());
				if(maxs != null) {
					for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> max:maxs) {
						addRequiredBinding(max);
					}
				}
			}
			else {
				Collection<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> nodes
									= iclValueToPatternNode.get(parameter);
				if(nodes != null) {
					for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:nodes) {
						addRequiredBinding(node);
					}
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		List<Object> query = constructQuery();
		Iterator<IndexEntry> entryIterator = index.getEntries(query);
		return new MLSDMIndexEntryIteratorMatchState(entryIterator);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!isExecutable()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		
		List<Object> query = constructQuery();
		return index.estimateCardinality(query);
	}

	/**
	 * Check whether the matching action can be executed, i.e.
	 * all required dependencies have already been bound.
	 * 
	 * @return	True if the action is executable.
	 */
	protected boolean isExecutable() {
		int parameterNumber = iclConstraint.getParameters().getList().size();
		
		/*
		 * Can only bind the last parameter if it is neither
		 * a range nor a value attribute and is unbound.
		 */
		ICLParameter last = iclConstraint.getParameters().getList().get(parameterNumber - 1);
		if(last.eClass() != ICLPackage.Literals.ICL_VARIABLE || getParameterValue((ICLValue) last) != null) {
			return false;
		}
		
		for(int i = 0; i < iclConstraint.getParameters().getList().size() - 1; i++) {
			ICLParameter parameter = iclConstraint.getParameters().getList().get(i);
			/*
			 * all parameters except for the last one have to be bound
			 * as this matching action should only match the last parameter
			 */
			if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
				if(getParameterValue(((ICLRange) parameter).getMin()) == null ||
						getParameterValue(((ICLRange) parameter).getMax()) == null) {
					return false;
				}
			}
			else {
				if(getParameterValue(((ICLValue) parameter)) == null) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Create an index query based on which dependencies
	 * are bound and which are not.
	 * 
	 * @return	An index query conforming to the current
	 * 			state of the search.
	 */
	protected List<Object> constructQuery() {
		List<Object> query = new ArrayList<Object>(iclConstraint.getParameters().getList().size() * 2);
		boolean foundUnbound = false;
		for(ICLParameter parameter:iclConstraint.getParameters().getList()) {
			if(foundUnbound) {
				query.add(Index.UNDEFINED_PARAMETER);
				query.add(Index.UNDEFINED_PARAMETER);
				continue;
			}
			
			Object minValue;
			Object maxValue;
			if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
				minValue = getParameterValue(((ICLRange) parameter).getMin());
				maxValue = getParameterValue(((ICLRange) parameter).getMax());
			}
			else {
				minValue = getParameterValue((ICLValue) parameter);
				maxValue = minValue;
			}
			
			if(minValue == null || maxValue == null) {
				query.add(Index.UNDEFINED_PARAMETER);
				query.add(Index.UNDEFINED_PARAMETER);
				foundUnbound = true;
			}
			else {
				query.add(minValue);
				query.add(maxValue);
			}
			
		}
		return query;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		//TODO this can probably be simplified for this particular case
		MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		transaction.setValid(false);
		
		if(!(matchState instanceof MLSDMIndexEntryIteratorMatchState)) {
			return transaction;
		}
		
		Iterator<IndexEntry> iterator = ((MLSDMIndexEntryIteratorMatchState) matchState).getEntryIterator();
		boolean valid = false;
		
		while(!valid && iterator.hasNext()) {
			transaction = new MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
			IndexEntry entry = iterator.next();
			
			/*
			 * If entry has already been used for matching skip to the next one.
			 */
			if(!matcher.checkMatchingHistory(this.getPatternConstraint(), entry)) {
				valid = false;
				transaction.setValid(valid);
				continue;
			}
			
			/*
			 * Check if the current entry is valid and create new bindings.
			 */
			List<ICLParameter> parameters = iclConstraint.getParameters().getList();
			Set<Object> newlyBoundObjects = new HashSet<Object>();
			valid = true;
			
			for(int i = 0; i < parameters.size(); i++) {
				ICLParameter parameter = parameters.get(i);
				Object indexValue = entry.getKey().get(i);

				if(parameter.eClass() == ICLPackage.Literals.ICL_VARIABLE &&
						iclValueToPatternNode.containsKey(parameter)) {
					for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:iclValueToPatternNode.get(parameter)) {
						if(patternNode.isBound()) {
							if(!checkValue((ICLValue) parameter, indexValue)) {
								valid = false;
								break;
							}
							else {
								continue;
							}
						}
					
						/*
						 * Create subtransactions for each previously unbound pattern node.
						 */
						if(matcher.instanceMatchesSPO(indexValue, patternNode.getSpo()) && !newlyBoundObjects.contains(indexValue)){
							MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> singleTransaction =
									new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
							singleTransaction.setValid(true);
							singleTransaction.setSPO(patternNode.getSpo());
							singleTransaction.setTargetInstance(indexValue);
							
							transaction.addSingleNodeTransaction(singleTransaction);
							transaction.addPatternNode(patternNode);
							
							newlyBoundObjects.add(indexValue);
						}
						else {
							valid = false;
							break;
						}
					}
				}
				else {
					/*
					 * Check already bound parameter.
					 */
					if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
						if(!checkRange((ICLRange) parameter, indexValue)) {
							valid = false;
							break;
						}
					}
					else {
						if(!checkValue((ICLValue) parameter, indexValue)) {
							valid = false;
							break;
						}
					}
				}
			}
			
			matcher.updateMatchingHistory(this.getPatternConstraint(), entry);
			transaction.setValid(valid);
		}
		
		return transaction;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "match staged\t(" + expression.getExpressionString() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNodeClones
				= new HashMap<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
		for(Entry<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> entry:iclValueToPatternNode.entrySet()) {
			List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> patternNodeClones = new ArrayList<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
			for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:entry.getValue()) {
				patternNodeClones.add((PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(pn));
			}
			iclValueToPatternNodeClones.put(entry.getKey(), patternNodeClones);
		}
		return new MLSDMIndexLastParameterMatchingAction(id, matcher, iclConstraint, expression, iclValueToPatternNodeClones);
	}
}
