package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckOnlyMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action which executes an explicit
 * check of a link. It can be executed once mappings have
 * been created for both source and target node of the link.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMLinkCheckMatchingAction extends MLSDMCheckMatchingAction {
	
	/*
	 * The link.
	 */
	protected StoryPatternLink link;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Target node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode;
	
	private CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction;
	
	public MLSDMLinkCheckMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher,
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint,
			StoryPatternLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode) {
		super(id, matcher, patternConstraint);
		this.link = link;
		this.sourceNode = sourceNode;
		this.targetNode = targetNode;
		addRequiredBinding(sourceNode);
		addRequiredBinding(targetNode);
		initializeTransaction();
	}

	private void initializeTransaction() {
		 transaction = new CheckOnlyMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(patternConstraint);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality()  {
		if(sourceNode.isBound() && targetNode.isBound()) {
			return PatternConstraint.CHECK_ONLY;
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	/**
	 * Perform the actual check of the link, i.e. check whether
	 * there is a matching link between the mappings for source
	 * and target node.
	 * 
	 * @return	True if a matching link exists.
	 */
	@SuppressWarnings("unchecked")
	protected boolean checkLink() {
		boolean success;
		EStructuralFeature feature = link.getFeature();
		AbstractStoryPatternObject sourceSPO = link.getSource();
		AbstractStoryPatternObject targetSPO = link.getTarget();
		
		EObject source = (EObject) matcher.getVariable(sourceSPO.getName()).getValue();
		EObject target = (EObject) matcher.getVariable(targetSPO.getName()).getValue();
		
		
		if(!feature.isMany()) {
			success = source.eGet(feature) == target;
		}
		else {
			List<? extends Object> instanceLinks = (List<? extends Object>) source.eGet(feature);
			switch(link.getLinkPositionConstraint()) {
				case BETWEEN :
					int index = instanceLinks.indexOf(target);
					success = index > 1 && index < instanceLinks.size() - 1;
					break;
				case UNCONSTRAINED :
					success = instanceLinks.contains(target);
					break;
				case FIRST :
					success = instanceLinks.size() > 0 && instanceLinks.get(0) == target;
					break;
				case LAST :
					success = instanceLinks.size() > 0 && instanceLinks.get(instanceLinks.size() - 1) == target;
					break;
				default :
					success = false;
					break;
			}				
		}
		
		if(success) {
			matcher.getNotificationEmitter().linkCheckSuccessful(link.getSource(), source, link, link.getTarget(), target, matcher.getVariablesScope(), matcher);
		}
		else {
			matcher.getNotificationEmitter().linkCheckFailed(link.getSource(), source, link, link.getTarget(), target, matcher.getVariablesScope(), matcher);
		}
		
		return success;
	}
	
	public StoryPatternLink getLink() {
		return link;
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		return new MLSDMCheckOnceMatchState();
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		transaction.setValid(((MLSDMCheckOnceMatchState) matchState).check() && checkLink());
		return transaction;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction#check()
	 */
	@Override
	public boolean check() {
		return checkBinding() && checkLink();
	}
	
	/**
	 * Check whether the pattern nodes required by this action's
	 * constraint still have a valid binding.
	 * 
	 * @return	True if valid bindings exist for all required
	 * 			pattern nodes.
	 */
	private boolean checkBinding() {
		return matcher.getVariable(sourceNode.getSpo().getName()) != null
				&& matcher.getVariable(targetNode.getSpo().getName()) != null;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "check\t\t(" + link.getSource().getName() + " -" + link.getFeature().getName() + "-> " + link.getTarget().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMLinkCheckMatchingAction(id, matcher, clonePatternConstraint, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(targetNode));
	}

}
