package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMReferenceIndex.MLSDMReferenceAdapterTuple;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchMultipleNodesMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching both source and target
 * node of a link. It can be executed as long as both source and
 * target node of the link have not yet been mapped.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMLinkLookupMatchingAction extends MLSDMMatchingAction {

	/**
	 * This is a match state specific to the link lookup matching
	 * action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private class MLSDMLinkLookupMatchState extends MatchState{
		
		/*
		 * Iterator of all matching instance links, represented by
		 * a tuple of source and target instance object.
		 */
		private Iterator<MLSDMReferenceAdapterTuple<EObject>> linkIterator;

		private MLSDMLinkLookupMatchState(Iterator<MLSDMReferenceAdapterTuple<EObject>> linkIterator2) {
			this.linkIterator = linkIterator2;
		}
		
	}

	/*
	 * The link.
	 */
	private StoryPatternLink link;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Target node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode;
	
	public MLSDMLinkLookupMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, StoryPatternLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetNode) {
		super(id, matcher);
		this.link = link;
		this.sourceNode = sourceNode;
		this.targetNode = targetNode;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@Override
	public MatchState createMatchState() {
		Iterator<MLSDMReferenceAdapterTuple<EObject>> linkIterator = matcher.getReferenceAdapter().getReferences((EReference) link.getFeature()).iterator();
		return new MLSDMLinkLookupMatchState(linkIterator);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(sourceNode.isBound() || targetNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else {
			return matcher.getReferenceAdapter().getReferences((EReference) link.getFeature()).size();
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchMultipleNodesMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint());
		boolean valid = false;
		
		if(matchState instanceof MLSDMLinkLookupMatchState) {
			Iterator<MLSDMReferenceAdapterTuple<EObject>> linkIterator = ((MLSDMLinkLookupMatchState)matchState).linkIterator;
			while(!valid && linkIterator.hasNext()) {
				MLSDMReferenceAdapterTuple<EObject> linkInstance = linkIterator.next();
				
				if(!matcher.checkMatchingHistory(getPatternConstraint(), linkInstance)) {
					continue;
				}
				
				/*
				 * Ensure isomorphism.
				 */
				valid = (sourceNode == targetNode || linkInstance.e1 != linkInstance.e2)
						&& matcher.instanceMatchesSPO(linkInstance.e1, sourceNode.getSpo())
						&& matcher.instanceMatchesSPO(linkInstance.e2, targetNode.getSpo())
						&& checkPositionConstraint(linkInstance.e1, linkInstance.e2);
				
				if(valid) {
					MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceTransaction =
							new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
					sourceTransaction.addPatternNode(sourceNode);
					sourceTransaction.setSPO(sourceNode.getSpo());
					sourceTransaction.setTargetInstance(linkInstance.e1);
					sourceTransaction.setValid(valid);
					
					MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> targetTransaction =
							new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
					targetTransaction.addPatternNode(targetNode);
					targetTransaction.setSPO(targetNode.getSpo());
					targetTransaction.setTargetInstance(linkInstance.e2);
					targetTransaction.setValid(valid);

					transaction.addPatternNode(sourceNode);
					transaction.addPatternNode(targetNode);
					transaction.addSingleNodeTransaction(sourceTransaction);
					transaction.addSingleNodeTransaction(targetTransaction);
				}
				
				matcher.updateMatchingHistory(getPatternConstraint(), linkInstance);
			}
		}
		
		transaction.setValid(valid);
		return transaction;
	}
	
	/**
	 * Check whether the given instance objects satisfy the position
	 * constraint imposed over this link.
	 * 
	 * @param e1
	 * 			the source instance
	 * @param e2
	 * 			the target instance
	 * @return	true if the position constraint is satisfied
	 */
	private boolean checkPositionConstraint(EObject e1, EObject e2) {
		List<?> instanceLinks;
		switch(link.getLinkPositionConstraint()) {
			case BETWEEN :
				instanceLinks = (List<?>)e1.eGet(link.getFeature());
				int index = instanceLinks.indexOf(e2);
				return index > 1 && index < instanceLinks.size() - 1;
			case UNCONSTRAINED :
				return true;
			case FIRST :
				instanceLinks = (List<?>)e1.eGet(link.getFeature());
				return instanceLinks.get(0) == e2;
			case LAST :
				instanceLinks = (List<?>)e1.eGet(link.getFeature());
				return instanceLinks.get(instanceLinks.size() - 1) == e2;
			default :
				return true;
		}
	}

	public EStructuralFeature getFeature() {
		return link.getFeature();
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getSourceNode() {
		return sourceNode;
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getTargetNode() {
		return targetNode;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "lookup\t\t(" + link.getSource().getName() + " -" + link.getFeature().getName() + "-> " + link.getTarget().getName() + ")";
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMLinkLookupMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(targetNode));
	}

}
