package de.mdelab.mlsdm.interpreter.searchModel.model.action;


import java.util.Map;

import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchSingleNodeMatchingTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelMatchingTransaction;

/**
 * This is a matching action for matching the value node of
 * a map entry link in the story pattern. It can be executed
 * once mappings have been created for both the source and
 * key node of the link. 
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMMapValueMatchingAction extends MLSDMMatchingAction {

	/**
	 * This is a match state specific to the map value matching
	 * action.
	 * 
	 * @author Matthias.Barkowsky
	 *
	 */
	private class MLSDMMapValueMatchState extends MatchState{
		
		/*
		 * Candidate instance binding for the value node.
		 */
		private Object valueInstance;
		
		/*
		 * Whether the action has been executed yet.
		 */
		private boolean matched;
		
		private MLSDMMapValueMatchState(Object valueInstance) {
			this.valueInstance = valueInstance;
			this.matched = false;
		}
		
		private Object getValueInstance() {
			return valueInstance;
		}

		private boolean isMatched() {
			return matched;
		}

		private void setMatched(boolean matched) {
			this.matched = matched;
		}
		
	}
	
	/*
	 * The map entry link.
	 */
	private MapEntryLink link;
	
	/*
	 * Source node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode;
	
	/*
	 * Value node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> valueNode;
	
	/*
	 * Key node of the link.
	 */
	private PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> keyNode;
	
	public MLSDMMapValueMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, MapEntryLink link,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> sourceNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> valueNode,
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> keyNode ) {
		super(id, matcher);
		this.link = link;
		this.sourceNode = sourceNode;
		this.valueNode = valueNode;
		this.keyNode = keyNode;
		addRequiredBinding(sourceNode);
		addRequiredBinding(keyNode);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#createMatchState()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public MatchState createMatchState() {
		if(!sourceNode.isBound() || !keyNode.isBound() || valueNode.isBound()) {
			return null;
		}
		else {
			EObject sourceInstance = (EObject) matcher.getVariable(sourceNode.getSpo().getName()).getValue();
			Object keyInstance = matcher.getVariable(keyNode.getSpo().getName()).getValue();
			Object valueInstance = ((EMap<Object, Object>) sourceInstance.eGet(link.getFeature())).get(keyInstance);
			return new MLSDMMapValueMatchState(valueInstance);
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		if(!sourceNode.isBound() || !keyNode.isBound() || valueNode.isBound()) {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
		else{
			return 1;
		}
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#match(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchState)
	 */
	@Override
	public SearchModelMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> match(
			MatchState matchState) {
		MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> transaction =
				new MatchSingleNodeMatchingTransaction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(getPatternConstraint(), matcher);
		
		EObject instanceObject = (EObject) matcher.getVariable(sourceNode.getSpo().getName()).getValue();
		matcher.getNotificationEmitter().traversingLink(link, link.getSource(), instanceObject, link.getTarget(), matcher.getVariablesScope(), matcher);
		
		if(!(matchState instanceof MLSDMMapValueMatchState)) {
			transaction.setValid(false);
		}
		else {
			boolean valid = false;
			MLSDMMapValueMatchState ms = (MLSDMMapValueMatchState) matchState;
			Object targetInstance = ms.getValueInstance();
			
			if(!ms.isMatched() && matcher.instanceMatchesSPO(targetInstance, valueNode.getSpo()) && matcher.checkMatchingHistory(getPatternConstraint(), targetInstance)) {
				valid = true;
				ms.setMatched(true);
			}
			
			if(valid) {
				transaction.setTargetInstance(targetInstance);
				transaction.setSPO(valueNode.getSpo());
			}
			else {
				matcher.getNotificationEmitter().storyPatternObjectNotBound(valueNode.getSpo(), matcher.getVariablesScope(),
						matcher);
			}
			
			transaction.setValid(valid);
			transaction.addPatternNode(valueNode);

			matcher.updateMatchingHistory(this.getPatternConstraint(), targetInstance);
		}
		return transaction;
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		return new MLSDMMapValueMatchingAction(id, matcher, link, (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(sourceNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(valueNode), (PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(keyNode));
	}
}
