package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

/**
 * This is an adapter which stores indices used for optimized
 * edge checks, reverse edge traversal, edge lookup and domain
 * matching. It automatically attaches to all instance objects in
 * the transitive closure of each added object to create and
 * maintain the indices. For maintenance, both EMF notifications
 * and notifications emitted by the story diagram interpreter are
 * considered.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMReferenceIndex extends MLSDMReferenceAdapter {

	/**
	 * @author Matthias.Barkowsky
	 *
	 * This is a simple tuple class for storing edges in the form
	 * of pairs of source and target node.
	 *
	 * @param <E>
	 */
	public static class MLSDMReferenceAdapterTuple<E> {
		
		public E e1;
		public E e2;
		
		public MLSDMReferenceAdapterTuple(E e1, E e2) {
			this.e1 = e1;
			this.e2 = e2;
		}
		
		@Override
		public int hashCode() {
			return e1.hashCode() ^ e2.hashCode();
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj == null || !(obj instanceof MLSDMReferenceAdapterTuple)) {
				return false;
			}
			MLSDMReferenceAdapterTuple<?> tuple = (MLSDMReferenceAdapterTuple<?>) obj;
			return e1 == tuple.e1 && e2 == tuple.e2;
		}
	}
	
	/*
	 * Map storing all instance objects registered at this
	 * reference adapter by their type.
	 */
	protected Map<EClass, Set<EObject>> instanceDomains;
	
	/*
	 * Map storing all objects referenced by a given source object
	 * in sets.
	 */
	protected Map<EObject, Set<EObject>[]> referencedObjects;
	
	/*
	 * Map storing all references to a given instance object.
	 */
	protected Map<EObject, Map<EReference, Set<EObject>>> inverseReferences;
	
	/*
	 * Map storing all instance links between registered objects
	 * by their reference type.
	 */
	protected Map<EReference, Set<MLSDMReferenceAdapterTuple<EObject>>> references;
	
	public MLSDMReferenceIndex() {
		instanceDomains = new HashMap<EClass, Set<EObject>>();
		referencedObjects = new HashMap<EObject, Set<EObject>[]>();
		inverseReferences = new HashMap<EObject, Map<EReference, Set<EObject>>>();
		references = new HashMap<EReference, Set<MLSDMReferenceAdapterTuple<EObject>>>();
	}

	/**
	 * Do add the given instance object to the registered objects
	 * and the domain map.
	 * 
	 * @param eObject
	 * 			The instance object.
	 */
	@Override
	protected void doAddEObject(EObject eObject) {
		EClass eClass = eObject.eClass();
		
		/*
		 * Add the object to the domain of all super types of
		 * the object's type.
		 */
		for(EClass type:getSuperTypes(eClass)) {
			if(!instanceDomains.containsKey(type)) {
				instanceDomains.put(type, new HashSet<EObject>());
			}
			
			instanceDomains.get(type).add(eObject);
		}

		/*
		 * Initialize the reference sets for the added object.
		 */
		initializeReferenceSets(eObject);
	}

	@SuppressWarnings("unchecked")
	private void initializeReferenceSets(EObject eObject) {
		if(!referencedObjects.containsKey(eObject)) {
			EClass eClass = eObject.eClass();
			Set<EObject>[] references = new Set[eClass.getFeatureCount()];
			for(int i = 0; i < references.length; i++) {
				references[i] = new HashSet<EObject>();
			}
			referencedObjects.put(eObject, references);
		}
		if(!inverseReferences.containsKey(eObject)) {
			inverseReferences.put(eObject, new HashMap<EReference, Set<EObject>>());
		}
	}

	/**
	 * Recursively collect all super types of a given type.
	 * 
	 * @param type
	 * 			The given type.
	 * @return	A set containing all super types of the given type
	 * 			and the given type itself.
	 */
	protected Set<EClass> getSuperTypes(EClass type) {
		Set<EClass> types = new HashSet<EClass>();
		types.add(type);
		types.addAll(type.getEAllSuperTypes());
		return types;
	}

	/**
	 * Remove an instance object from the adapter. This does not
	 * include the removal of incoming and outgoing instance links.
	 * 
	 * @param eObject
	 * 			The instance object to remove.
	 */
	@Override
	protected void doRemoveEObject(EObject eObject) {
		for(EClass type:getSuperTypes(eObject.eClass())) {
			if(instanceDomains.containsKey(type)) {
				instanceDomains.get(type).remove(eObject);
			}
		}
		
		referencedObjects.remove(eObject);
		inverseReferences.remove(eObject);
	}

	/**
	 * Update internal data structures of this adapter with a new
	 * given instance link.
	 * 
	 * @param source
	 * 			The source instance object of the link.
	 * @param reference
	 * 			The type of the link.
	 * @param target
	 * 			The target instance object of the link.
	 */
	@Override
	protected void doAddReference(EObject source, EReference reference, EObject target) {
		/*
		 * Ensure proper reference sets are initialized for source and target.
		 */
		initializeReferenceSets(source);
		initializeReferenceSets(target);

		/*
		 * Update reference set.
		 */
		referencedObjects.get(source)[reference.getFeatureID()].add(target);
		
		/*
		 * Update inverse references.
		 */
		Set<EObject> inverseReference = inverseReferences.get(target).get(reference);
		if(inverseReference == null) {
			inverseReference = new HashSet<EObject>();
			inverseReferences.get(target).put(reference, inverseReference);
		}
		inverseReference.add(source);
		
		/*
		 * Update reference type map.
		 */
		if(!references.containsKey(reference)) {
			references.put(reference, new HashSet<MLSDMReferenceAdapterTuple<EObject>>());
		}
		MLSDMReferenceAdapterTuple<EObject> tuple = new MLSDMReferenceAdapterTuple<EObject>(source, target);
		references.get(reference).add(tuple);
	}
	
	/**
	 * Remove a given instance link from the data structures of this
	 * adapter.
	 * 
	 * @param source
	 * 			The source instance object of the link.
	 * @param reference
	 * 			The type of the link.
	 * @param target
	 * 			The target instance object of the link.
	 */
	@Override
	protected void doRemoveReference(EObject source, EReference reference, EObject target) {
		if(referencedObjects.containsKey(source)) {
			referencedObjects.get(source)[reference.getFeatureID()].remove(target);
		}
		
		if(inverseReferences.containsKey(target) && inverseReferences.get(target).containsKey(reference)) {
			inverseReferences.get(target).get(reference).remove(source);
		}
		
		if(references.containsKey(reference)) {
			MLSDMReferenceAdapterTuple<EObject> tuple = new MLSDMReferenceAdapterTuple<EObject>(source, target);
			references.get(reference).remove(tuple);
		}
	}
	
	/**
	 * Checks if a given instance link exists in the registered model.
	 * 
	 * @param source
	 * 			The source instance object of the link.
	 * @param reference
	 * 			The type of the link.
	 * @param target
	 * 			The target instance object of the link.
	 * @return	True if a link exists in the model that matches the
	 * 			given link.
	 */
	public boolean checkReference(EObject source, EReference reference, EObject target) {
		return referencedObjects.containsKey(source) ? referencedObjects.get(source)[reference.getFeatureID()].contains(target) : false;
	}

	/**
	 * Retrieve all incoming links of a given type to a given instance
	 * object.
	 * 
	 * @param eObject
	 * 			The instance object.
	 * @param reference
	 * 			The type of the incoming links.
	 * @return	A collection containing all objects referencing the
	 * 			given instance object by a link of the given type.
	 */
	public Collection<EObject> getInverseReferences(EObject eObject, EReference reference){
		if(inverseReferences.containsKey(eObject) && inverseReferences.get(eObject).containsKey(reference)) {
			return inverseReferences.get(eObject).get(reference);
		}
		else {
			return Collections.emptySet();
		}
	}

	public Map<EReference, Set<EObject>> getInverseReferences(EObject eObject) {
		if(inverseReferences.containsKey(eObject)) {
			return inverseReferences.get(eObject);
		}
		else {
			return Collections.emptyMap();
		}
	}

	/**
	 * Retrieve all references of a given type.
	 * 
	 * @param feature
	 * 			The type of the references to retrieve.
	 * @return	A collection of (source, target) tuples representing
	 * 			one link of the given type each.
	 */
	@SuppressWarnings("unchecked")
	public Collection<MLSDMReferenceAdapterTuple<EObject>> getReferences(EReference feature) {
		if(references.containsKey(feature)) {
			return references.get(feature);
		}
		else {
			return (Collection<MLSDMReferenceAdapterTuple<EObject>>) Collections.EMPTY_SET;
		}
	}
	
	public Map<EReference, Set<MLSDMReferenceAdapterTuple<EObject>>> getReferences() {
		return references;
	}

	public Map<EClass, Set<EObject>> getInstanceDomains() {
		return instanceDomains;
	}

	/**
	 * Retrieve all instance objects with a given type.
	 * 
	 * @param eClass
	 * 			The type.
	 * @return	A collection containing all objects of the given
	 * 			type registered at this adapter. This includes
	 * 			instances of subclasses of the given type.
	 */
	@SuppressWarnings("unchecked")
	public Collection<EObject> getDomain(EClass eClass){
		if(!instanceDomains.containsKey(eClass)) {
			return (Collection<EObject>) Collections.EMPTY_SET;
		}
		return instanceDomains.get(eClass);
	}
	
}
