package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlcallactions.ActivityCallAction;
import de.mdelab.mlcallactions.BinaryOperationAction;
import de.mdelab.mlcallactions.CallAction;
import de.mdelab.mlcallactions.CallActionExpression;
import de.mdelab.mlcallactions.CallActionParameter;
import de.mdelab.mlcallactions.EOperationCallAction;
import de.mdelab.mlcallactions.MethodCallAction;
import de.mdelab.mlcallactions.MlcallactionsPackage;
import de.mdelab.mlcallactions.NewObjectAction;
import de.mdelab.mlcallactions.ParameterizedCallAction;
import de.mdelab.mlcallactions.UnaryOperationAction;
import de.mdelab.mlcallactions.VariableDeclarationAction;
import de.mdelab.mlcallactions.VariableReferenceAction;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzerManager;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.FeatureAccess;

/**
 * This is an expression analyzer implementation for the call actions
 * language.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMCallActionsExpressionAnalyzer extends MLSDMExpressionAnalyzer {

	public MLSDMCallActionsExpressionAnalyzer(){
		this(null);	//TODO factory
	}
	
	public MLSDMCallActionsExpressionAnalyzer(
			ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> analyzerManager) {
		super(analyzerManager);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#getVariableNames(java.lang.Object)
	 */
	@Override
	public Collection<String> getVariableNames(MLExpression expression) {
		if(!(expression instanceof CallActionExpression)) {
			return null;
		}
		
		Set<String> variableNames = new HashSet<String>();
		CallActionExpression cAExpression = (CallActionExpression) expression;
		for(CallAction callAction:cAExpression.getCallActions()) {
			variableNames.addAll(getVariableNames(callAction));
		}
		return variableNames;
	}

	/**
	 * Retrieve the names of all variables referenced in a call action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The call action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNames(CallAction callAction) {
		Collection<String> variableNames;
		EClass callActionEClass = callAction.eClass();

		if (callActionEClass == MlcallactionsPackage.Literals.LITERAL_DECLARATION_ACTION)
		{
			variableNames = Collections.emptySet();
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.ACTIVITY_CALL_ACTION)
		{
			variableNames = getVariableNamesParametrizedAction((ActivityCallAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.NEW_OBJECT_ACTION)
		{
			variableNames = getVariableNamesParametrizedAction((NewObjectAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.VARIABLE_DECLARATION_ACTION)
		{
			variableNames = getVariableNamesVariableDeclaration((VariableDeclarationAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.UNARY_OPERATION_ACTION)
		{
			variableNames = getVariableNamesUnaryOperation((UnaryOperationAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.BINARY_OPERATION_ACTION)
		{
			variableNames = getVariableNamesBinaryOperation((BinaryOperationAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.NULL_VALUE_ACTION)
		{
			variableNames = Collections.emptySet();
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.METHOD_CALL_ACTION)
		{
			variableNames = getVariableNamesMethodCall((MethodCallAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.EOPERATION_CALL_ACTION)
		{
			variableNames = getVariableNamesEOperationCall((EOperationCallAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.VARIABLE_REFERENCE_ACTION)
		{
			variableNames = getVariableNamesVariableReference((VariableReferenceAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.EOBJECT_REFERENCE_ACTION)
		{
			variableNames = Collections.emptySet();
		}
		else
		{
			throw new UnsupportedOperationException();
		}
		return variableNames;
	}

	/**
	 * Retrieve the names of all variables referenced in a parametrized call action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The parametrized call action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNamesParametrizedAction(ParameterizedCallAction callAction) {
		Set<String> variableNames = new HashSet<String>();
		for(CallActionParameter parameter:callAction.getParameters()) {
			variableNames.addAll(analyzerManager.getVariableNames(parameter.getValueExpression()));
		}
		return variableNames;
	}

	/**
	 * Retrieve the names of all variables referenced in a unary call action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The unary call action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNamesUnaryOperation(UnaryOperationAction callAction) {
		return analyzerManager.getVariableNames(callAction.getOperand());
	}

	/**
	 * Retrieve the names of all variables referenced in a variable declaration action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The variable declaration action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNamesVariableDeclaration(VariableDeclarationAction callAction) {
		return analyzerManager.getVariableNames(callAction.getValueExpression());
	}

	/**
	 * Retrieve the names of all variables referenced in a binary call action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The binary call action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNamesBinaryOperation(BinaryOperationAction callAction) {
		Set<String> variableNames = new HashSet<String>();
		variableNames.addAll(analyzerManager.getVariableNames(callAction.getOperand1()));
		variableNames.addAll(analyzerManager.getVariableNames(callAction.getOperand2()));
		return variableNames;
	}

	/**
	 * Retrieve the names of all variables referenced in a method call action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The method call action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNamesMethodCall(MethodCallAction callAction) {
		Set<String> variableNames = new HashSet<String>();
		variableNames.addAll(getVariableNamesParametrizedAction(callAction));
		variableNames.addAll(analyzerManager.getVariableNames(callAction.getThisParameterValue()));
		return variableNames;
	}

	/**
	 * Retrieve the names of all variables referenced in a EOperation call action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The EOperation call action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNamesEOperationCall(EOperationCallAction callAction) {
		Set<String> variableNames = new HashSet<String>();
		variableNames.addAll(getVariableNamesParametrizedAction(callAction));
		variableNames.addAll(analyzerManager.getVariableNames(callAction.getThisParameterValue()));
		return variableNames;
	}

	/**
	 * Retrieve the names of all variables referenced in a variable reference action
	 * and its nested actions.
	 * 
	 * @param callAction
	 * 			The variable reference action.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	private Collection<String> getVariableNamesVariableReference(VariableReferenceAction callAction) {
		Set<String> variableNames = new HashSet<String>();
		variableNames.add(callAction.getName());
		return variableNames;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#parseAST(java.lang.Object)
	 */
	@Override
	public Object parseAST(MLExpression expression, EClassifier contextClassifier) {
		return expression;
	}

	@Override
	public Collection<FeatureAccess<EStructuralFeature>> getFeatureAccesses(
			MLExpression expression, EClassifier contextClassifier) {
		if(!(expression instanceof CallActionExpression)) {
			return null;
		}
		
		Set<FeatureAccess<EStructuralFeature>> accesses = new HashSet<FeatureAccess<EStructuralFeature>>();
		CallActionExpression cAExpression = (CallActionExpression) expression;
		for(CallAction callAction:cAExpression.getCallActions()) {
			accesses.addAll(getFeatureAccesses(callAction));
		}
		return accesses;
	}

	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccesses(CallAction callAction) {
		Collection<FeatureAccess<EStructuralFeature>> accesses;
		EClass callActionEClass = callAction.eClass();

		if (callActionEClass == MlcallactionsPackage.Literals.LITERAL_DECLARATION_ACTION)
		{
			accesses = Collections.emptySet();
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.ACTIVITY_CALL_ACTION)
		{
			accesses = getFeatureAccessesParametrizedAction((ActivityCallAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.NEW_OBJECT_ACTION)
		{
			accesses = getFeatureAccessesParametrizedAction((NewObjectAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.VARIABLE_DECLARATION_ACTION)
		{
			accesses = getFeatureAccessesVariableDeclaration((VariableDeclarationAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.UNARY_OPERATION_ACTION)
		{
			accesses = getFeatureAccessesUnaryOperation((UnaryOperationAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.BINARY_OPERATION_ACTION)
		{
			accesses = getFeatureAccessesBinaryOperation((BinaryOperationAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.NULL_VALUE_ACTION)
		{
			accesses = Collections.emptySet();
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.METHOD_CALL_ACTION)
		{
			accesses = getFeatureAccessesMethodCall((MethodCallAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.EOPERATION_CALL_ACTION)
		{
			accesses = getFeatureAccessesEOperationCall((EOperationCallAction) callAction);
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.VARIABLE_REFERENCE_ACTION)
		{
			accesses = Collections.emptySet();
		}
		else if (callActionEClass == MlcallactionsPackage.Literals.EOBJECT_REFERENCE_ACTION)
		{
			accesses = Collections.emptySet();
		}
		else
		{
			throw new UnsupportedOperationException();
		}
		return accesses;
	}

	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccessesEOperationCall(
			EOperationCallAction callAction) {
		Set<FeatureAccess<EStructuralFeature>> accesses = new HashSet<FeatureAccess<EStructuralFeature>>();
		accesses.addAll(getFeatureAccessesParametrizedAction(callAction));
		accesses.addAll(analyzerManager.getFeatureAccesses(callAction.getThisParameterValue()));
		return accesses;
	}

	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccessesMethodCall(
			MethodCallAction callAction) {
		Set<FeatureAccess<EStructuralFeature>> accesses = new HashSet<FeatureAccess<EStructuralFeature>>();
		accesses.addAll(getFeatureAccessesParametrizedAction(callAction));
		accesses.addAll(analyzerManager.getFeatureAccesses(callAction.getThisParameterValue()));
		return accesses;
	}

	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccessesBinaryOperation(
			BinaryOperationAction callAction) {
		Set<FeatureAccess<EStructuralFeature>> accesses = new HashSet<FeatureAccess<EStructuralFeature>>();
		accesses.addAll(analyzerManager.getFeatureAccesses(callAction.getOperand1()));
		accesses.addAll(analyzerManager.getFeatureAccesses(callAction.getOperand2()));
		return accesses;
	}

	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccessesUnaryOperation(
			UnaryOperationAction callAction) {
		return analyzerManager.getFeatureAccesses(callAction.getOperand());
	}

	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccessesVariableDeclaration(
			VariableDeclarationAction callAction) {
		return analyzerManager.getFeatureAccesses(callAction.getValueExpression());
	}

	private Collection<FeatureAccess<EStructuralFeature>> getFeatureAccessesParametrizedAction(ParameterizedCallAction callAction) {
		Set<FeatureAccess<EStructuralFeature>> variableNames = new HashSet<FeatureAccess<EStructuralFeature>>();
		for(CallActionParameter parameter:callAction.getParameters()) {
			variableNames.addAll(analyzerManager.getFeatureAccesses(parameter.getValueExpression()));
		}
		return variableNames;
	}

}
