package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.heap;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * This is a binary (min-)heap implementation which performs
 * operations such as insert and pop in the form of transactions
 * that can be rolled back to undo the respective operation.
 * 
 * @author Matthias.Barkowsky
 *
 * @param <T>
 */
public class MLSDMBinaryHeapWithRollBack<T> {
	
	/*
	 * Raw data in the heap.
	 */
	protected MLSDMHeapEntry<T>[] elements;
	
	/*
	 * Size of the heap.
	 */
	private int size;
	
	/*
	 * Ongoing heap operation. It caches all changes to the
	 * heap data structure and composes them into a single
	 * transaction when finishCurrentOperation is called.
	 */
	private MLSDMHeapOperation<T> currentOperation;
	
	/*
	 * Index allowing the retrieval of the position of a certain
	 * element in the heap.
	 */
	private Map<T, Integer> positions;
	
	public MLSDMBinaryHeapWithRollBack() {
		this(16);
	}
	
	@SuppressWarnings("unchecked")
	public MLSDMBinaryHeapWithRollBack(int initialSize) {
		this.elements = new MLSDMHeapEntry[initialSize];
		this.size = 0;
		this.currentOperation = new MLSDMHeapOperation<T>();
		this.positions = new HashMap<T, Integer>();
	}
	
	/**
	 * Insert the value under the given key.
	 * 
	 * @param key
	 * 			The key.
	 * @param value
	 * 			The value.
	 */
	public void insert(int key, T value) {
		int index = size;
		assertSize(index + 1);
		elements[index] = new MLSDMHeapEntry<T>(Integer.MAX_VALUE, value);
		positions.put(value, index);
		decreaseKey(index, key);
		grow();
	}
	
	/**
	 * Ensure the heap is big enough to store the given number of
	 * elements.
	 * 
	 * @param i
	 * 			The number of elements.
	 */
	private void assertSize(int i) {
		if(i >= elements.length) {
			elements = Arrays.copyOf(elements, elements.length * 2);
		}
	}

	/**
	 * Extract the element stored under the smallest key in the
	 * heap.
	 * 
	 * @return	The extracted element.
	 */
	public T extractMin() {
		if(size < 1) {
			return null;
		}
		
		MLSDMHeapEntry<T> entry = elements[0];
		int lastIndex = size - 1;
		swap(0, lastIndex);
		shrink();
		
		if(0 != lastIndex) {
			heapify(0);
		}
		return entry.value;
	}

	/**
	 * Decrement the heap size.
	 */
	private void shrink() {
		size--;
		currentOperation.sizeChange--;
	}
	
	/**
	 * Increment the heap size.
	 */
	private void grow() {
		size++;
		currentOperation.sizeChange++;
	}

	/**
	 * Retrieve the element stored under the smallest key in
	 * the heap without removing it from the data structure.
	 * 
	 * @return	The elementstored under the smallest key in the
	 * 			heap.
	 */
	public T getMin() {
		if(size < 1) {
			return null;
		}
		
		return elements[0].value;
	}

	/**
	 * Decrease the key under which a given element is
	 * stored in the heap and update the internal data
	 * structure accordingly.
	 * 
	 * @param value
	 * 			The element whose key is to be decreased.
	 * @param l
	 * 			The new, decreased key.
	 */
	public void decreaseKey(T value, double l) {
		int index = find(value);
		decreaseKey(index, l);
	}
	
	/**
	 * Retrieve the position of the given element in the
	 * underlying array.
	 * 
	 * @param value
	 * 			The element.
	 * @return	The index of the element.
	 */
	private int find(T value) {
		return positions.get(value);
	}

	/**
	 * Decrease the key of the element at the given index
	 * to the given new key and update the internal data
	 * structure accordingly.
	 * 
	 * @param index
	 * 			The index.
	 * @param newKey
	 * 			The new, decreased key.
	 */
	public void decreaseKey(int index, double newKey) {
		changeKey(index, newKey);
		int parent = parent(index);
		while(index > 0 && elements[index].key < elements[parent].key) {
			swap(index, parent);
			index = parent;
			parent = parent(index);
		}
	}
	
	/**
	 * Set the key of the heap entry at the given index to
	 * the given key.
	 * 
	 * @param index
	 * 			The index.
	 * @param newKey
	 * 			The new key.
	 */
	private void changeKey(int index, double newKey) {
		currentOperation.keyChanges.add(elements[index]);
		currentOperation.oldKeys.add(elements[index].key);
		elements[index].key = newKey;
	}

	/**
	 * Ensure the heap property is satisfied by the element
	 * at the given index.
	 * 
	 * @param index
	 */
	protected void heapify(int index) {
		int i = index;
		while(true) {
			int min = i;
			int left = left(i);
			int right = right(i);
			
			if(left < size && elements[left].key < elements[min].key) {
				min = left;
			}
			if(right < size && elements[right].key < elements[min].key) {
				min = right;
			}
			if(min == i) {
				break;
			}
			
			swap(i, min);
			i = min;
		}
	}

	public int size() {
		return size ;
	}
	
	/**
	 * Swap the heap entries at the given positions.
	 * 
	 * @param i1
	 * 			The first position.
	 * @param i2
	 * 			The second position.
	 */
	protected void swap(int i1, int i2) {
		currentOperation.swap1.add(i1);
		currentOperation.swap2.add(i2);
		doSwap(i1, i2);
	}

	/**
	 * Actually perform the swap of the heap entries
	 * at the given positions.
	 * 
	 * @param i1
	 * 			The first position.
	 * @param i2
	 * 			The second position.
	 */
	private void doSwap(int i1, int i2) {
		positions.put(elements[i1].value, i2);
		positions.put(elements[i2].value, i1);
		MLSDMHeapEntry<T> tmp = elements[i1];
		elements[i1] = elements[i2];
		elements[i2] = tmp;
	}

	/**
	 * Retrieve the index of the parent element of the
	 * element at the given index.
	 * 
	 * @param index
	 * 			The index.
	 * @return	The parent's index.
	 */
	private int parent(int index) {
		return (index - 1) / 2;
	}

	/**
	 * Retrieve the index of the left child element of
	 * the element at the given index.
	 * 
	 * @param index
	 * 			The index.
	 * @return	The left child's index.
	 */
	private int left(int index) {
		return (2 * index) + 1;
	}
	
	/**
	 * Retrieve the index of the right child element of
	 * the element at the given index.
	 * 
	 * @param index
	 * 			The index.
	 * @return	The right child's index.
	 */
	private int right(int index) {
		return (2 * index) + 2;
	}

	/**
	 * Complete the current heap operation and wrap it
	 * into a transaction.
	 * 
	 * @return	The current heap operation.
	 */
	public MLSDMHeapOperation<T> finishCurrentOperation() {
		MLSDMHeapOperation<T> op = currentOperation;
		currentOperation = new MLSDMHeapOperation<T>();
		return op;
	}
	
	/**
	 * Undo a given heap operation.
	 * 
	 * @param operation
	 * 			The heap operation.
	 */
	public void rollBack(MLSDMHeapOperation<T> operation) {
		for(int i = operation.swap1.size() - 1; i >= 0; i--) {
			doSwap(operation.swap1.get(i), operation.swap2.get(i));
		}
		
		for(int i = operation.keyChanges.size() - 1; i >= 0; i--) {
			operation.keyChanges.get(i).key = operation.oldKeys.get(i);
		}
		
		size -= operation.sizeChange;
	}

	/**
	 * Retrieve the element stored at the given index
	 * in the underlying array data structure.
	 * 
	 * @param index
	 * 			The index.
	 * @return	The element at the given index.
	 */
	public T get(int index) {
		if(index >= size) {
			return null;
		}
		else{
			return elements[index].value;
		}
	}

}
