package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.heap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMCardinalityMatchingCostStrategy;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMMatchingCostStrategy;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy;

/**
 * This is a pattern constraint selection strategy based on a cost
 * model for pattern constraint checks. All unchecked pattern constraints
 * are stored in a heap data structure alongside their matching cost to
 * allow an efficient retrieval of the cheapest pattern constraint to
 * handle in each matching step.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMHeapBasedSelectionStrategy extends OrderProducingSelectionStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {

	/*
	 * Number of successive explicit constraint checks after
	 * the last pattern node has been mapped.
	 */
	protected int lastCheck;
	
	/*
	 * Number of successive explicit constraint checks of optional
	 * constraints after the last pattern node has been mapped.
	 */
	protected int lastCheckOptional;
	
	/*
	 * Used heap data structure.
	 */
	protected MLSDMBinaryHeapWithRollBack<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> heap;
	
	/*
	 * Performed heap operations.
	 */
	protected List<MLSDMHeapOperation<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> updates;
	
	/*
	 * Additional heap data structure for optional constraints.
	 */
	protected MLSDMBinaryHeapWithRollBack<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> heapOptional;
	
	/*
	 * Performed heap operations on the heap for optional
	 * constraints.
	 */
	protected List<MLSDMHeapOperation<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> updatesOptional;

	/*
	 * List of heap operations corresponding to pattern constraint
	 * pop operations.
	 */
	protected List<MLSDMHeapOperation<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> pops;

	/*
	 * Strategy encoding the used cost model.
	 */
	protected MLSDMMatchingCostStrategy costStrategy;
	
	public MLSDMHeapBasedSelectionStrategy() {
		this.costStrategy = new MLSDMCardinalityMatchingCostStrategy();
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy#doUpdate(java.util.Collection)
	 */
	@Override
	protected void doUpdate(
			Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> list) {
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode:list) {
			for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint:patternNode.getDependantPatternConstraints()) {
				if(patternConstraint.isActive()) {
					continue;
				}
				
				if(!patternConstraint.isOptional()) {
					heap.decreaseKey(patternConstraint, costStrategy.calculateMatchingCost(patternConstraint));
				}
				else {
					heapOptional.decreaseKey(patternConstraint, costStrategy.calculateMatchingCost(patternConstraint));
				}
			}
		}

		updates.add(heap.finishCurrentOperation());
		updatesOptional.add(heapOptional.finishCurrentOperation());
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy#doInitialize()
	 */
	@Override
	protected void doInitialize() {
		lastCheck = -1;
		lastCheckOptional = -1;
		
		heap = new MLSDMBinaryHeapWithRollBack<PatternConstraint<StoryPattern,AbstractStoryPatternObject,AbstractStoryPatternLink,EClassifier,EStructuralFeature,MLExpression>>(searchModel.getPatternConstraints().size());
		heapOptional = new MLSDMBinaryHeapWithRollBack<PatternConstraint<StoryPattern,AbstractStoryPatternObject,AbstractStoryPatternLink,EClassifier,EStructuralFeature,MLExpression>>(searchModel.getPatternConstraints().size());
		updates = new ArrayList<MLSDMHeapOperation<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
		updatesOptional = new ArrayList<MLSDMHeapOperation<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
		
		pops = new ArrayList<MLSDMHeapOperation<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
		
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint:searchModel.getPatternConstraints()) {
			if(!patternConstraint.isOptional()) {
				heap.insert(Integer.MAX_VALUE, patternConstraint);
			}
			else {
				heapOptional.insert(Integer.MAX_VALUE, patternConstraint);
			}
		}

		heap.finishCurrentOperation();
		heapOptional.finishCurrentOperation();
		
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint:searchModel.getPatternConstraints()) {
			if(patternConstraint.isActive()) {
				continue;
			}
			if(!patternConstraint.isOptional()) {
				heap.decreaseKey(patternConstraint, costStrategy.calculateMatchingCost(patternConstraint));
			}
			else {
				heapOptional.decreaseKey(patternConstraint, costStrategy.calculateMatchingCost(patternConstraint));
			}
		}

		updates.add(heap.finishCurrentOperation());
		updatesOptional.add(heap.finishCurrentOperation());
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy#doRollBackLastPop()
	 */
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> doRollBackLastPop() {
		PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> lastPop = matchingOrder.remove(matchingOrder.size() - 1);
		
		if(!lastPop.isOptional()) {
			if(lastCheck >= 0) {
				lastCheck--;
			}
			else{
				heap.rollBack(pops.remove(pops.size() - 1));
			}
		}
		else{
			if(lastCheckOptional >= 0) {
				lastCheckOptional--;
			}
			else{
				heapOptional.rollBack(pops.remove(pops.size() - 1));
			}
		}
		
		return lastPop;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy#doRollBackLastUpdate()
	 */
	@Override
	protected void doRollBackLastUpdate() {
		heap.rollBack(updates.remove(updates.size() - 1));
		heapOptional.rollBack(updatesOptional.remove(updates.size() - 1));
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy#selectNextConstraint()
	 */
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> selectNextConstraint() {
		PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node;
		
		if(searchModel.getBoundPatternNodeNumber() == searchModel.getPatternNodes().size()) {
			node = heap.get(lastCheck + 1);
			if(node != null) {
				lastCheck++;
			}
			else {
				node = heapOptional.get(lastCheckOptional + 1);
				if(node != null) {
					lastCheckOptional++;
				}
			}
			return node;
		}
		
		node = heap.getMin();
		if(node != null && node.getCardinality() != PatternConstraint.MATCHING_NOT_POSSIBLE) {
			heap.extractMin();
			pops.add(heap.finishCurrentOperation());
		}
		else {
			node = heapOptional.getMin();
			if(node != null) {
				heapOptional.extractMin();
				pops.add(heapOptional.finishCurrentOperation());
			}
		}
		return node;
	}

}
