package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Pattern;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMDomainMatchMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkLookupMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseReverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.mlindices.StagedIndex;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public abstract class MLSDMFixedCostCalculator
		extends
			MLSDMConstraintSensitiveCostCalculator {
	
	public static final String DELIMITER = Pattern.quote(";");
	
	protected Map<Object, Double> domainCosts;
	protected Map<Object, Double> lookupCosts;
	protected Map<Object, Double> traversalCosts;
	protected Map<Object, Double> reverseTraversalCosts;
	protected Map<Object, Double> filters;
	
	public MLSDMFixedCostCalculator(
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			MLSDMMetadataIndex metadataAdapter, CostModel costModel,
			VariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super(searchModel, metadataAdapter, costModel, variablesScope);
	}
	
	@Override
	public Adornment[] calculateCotinuationCosts() {
		loadCostTables();
		
		return super.calculateCotinuationCosts();
	}

	protected void loadCostTables() {
		domainCosts = new HashMap<Object, Double>();
		lookupCosts = new HashMap<Object, Double>();
		traversalCosts = new HashMap<Object, Double>();
		reverseTraversalCosts = new HashMap<Object, Double>();
		filters = new HashMap<Object, Double>();
		
		try {
			loadEdgeCosts();
			loadDomainCosts();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void loadDomainCosts() throws IOException {
		FileReader nodeFR = new FileReader(getNodeFile());
		BufferedReader nodeBR = new BufferedReader(nodeFR);
		
		String line;
		while((line = nodeBR.readLine()) != null) {
			String[] tokens = line.split(DELIMITER);
			String className = tokens[0];
			EClass eClass = getEClass(className);
			
			domainCosts.put(eClass, Double.parseDouble(tokens[1]));
		}
		
		nodeBR.close();
		nodeFR.close();
	}

	private void loadEdgeCosts() throws IOException {
		FileReader edgeFR = new FileReader(getEdgeFile());
		BufferedReader edgeBR = new BufferedReader(edgeFR);
		
		String line;
		while((line = edgeBR.readLine()) != null) {
			String[] tokens = line.split(DELIMITER);
			String[] feature = tokens[0].split(Pattern.quote("."));
			
			EReference reference = getEReference(feature[0], feature[1]);
			
			traversalCosts.put(reference, Double.parseDouble(tokens[1]));
			reverseTraversalCosts.put(reference, Double.parseDouble(tokens[2]));
			lookupCosts.put(reference, Double.parseDouble(tokens[3]));
		}
		
		edgeBR.close();
		edgeFR.close();
	}

	private EClass getEClass(String className) {
		for(EPackage pkg:getPackages()) {
			EClass c = (EClass) pkg.getEClassifier(className);
			if(c != null) {
				return c;
			}
		}
		return null;
	}

	private EReference getEReference(String className, String referenceName) {
		for(EPackage pkg:getPackages()) {
			EClass c = (EClass) pkg.getEClassifier(className);
			if(c != null) {
				for(EReference reference:c.getEAllReferences()) {
					if(reference.getName().equals(referenceName)) {
						return reference;
					}
				}
			}
		}
		return null;
	}

	@Override
	public double estimateCost(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		if(ma instanceof MLSDMDomainMatchMatchingAction) {
			return domainCosts.get((EClass) ma.getPatternConstraint().getConstraint());
		}
		else if(ma instanceof MLSDMLinkTraverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseMatchingAction) ma).getFeature();
			return traversalCosts.get(reference);
		}
		else if(ma instanceof MLSDMLinkTraverseReverseMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkTraverseReverseMatchingAction) ma).getFeature();
			return reverseTraversalCosts.get(reference);
		}
		else if(ma instanceof MLSDMLinkLookupMatchingAction) {
			EReference reference = (EReference) ((MLSDMLinkLookupMatchingAction) ma).getFeature();
			return lookupCosts.get(reference);
		}
		else {
			return PatternConstraint.MATCHING_NOT_POSSIBLE;
		}
	}

	@Override
	protected double estimateStagedIndexCost(StagedIndex index, int level) {
		return index.getDepth() - level;
	}

	@Override
	public double getFiltering(
			CheckMatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> explicitCheckAction) {
		Object constraint = explicitCheckAction.getPatternConstraint().getConstraint();
		return filters.containsKey(constraint) ? filters.get(constraint) : 1;
	}
	
	protected abstract String getNodeFile();

	protected abstract String getEdgeFile();

	protected abstract Iterable<EPackage> getPackages();
}
